!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   at_remd_str_mod
!> @brief   structure of replica information
!! @authors Takaharu Mori (TM)
! 
!  (c) Copyright 2014 RIKEN. All rights reserved.
!
!--------1---------2---------3---------4---------5---------6---------7---------8

#ifdef HAVE_CONFIG_H
#include "../config.h"
#endif

module at_remd_str_mod

  use messages_mod
  use constants_mod

  implicit none
  private

  ! structures
  type, public :: s_remd
    integer                           :: total_nreplicas
    integer                           :: ncycles
    integer                           :: dimension
    integer                           :: exchange_period
    integer                           :: iseed
    logical                           :: equilibration_only
    integer,              allocatable :: types(:)
    integer,              allocatable :: nreplicas(:)
    integer,              allocatable :: repid2parmsetid(:)
    integer,              allocatable :: repid2parmsetid_ref(:)
    integer,              allocatable :: parmsetid2repid(:)
    integer,              allocatable :: parmidsets(:,:)
    integer,              allocatable :: num_criteria(:,:,:)
    integer,              allocatable :: num_exchanges(:,:,:)
    integer,              allocatable :: rest_function(:)
    integer,              allocatable :: umbrid2funcid(:)
    integer,              allocatable :: iparameters(:,:)
    integer,              allocatable :: umbrid2numfuncs(:)
    integer,              allocatable :: umbrid2funclist(:,:)
    real(wp),             allocatable :: dparameters(:,:)
    real(wp),             allocatable :: rest_constants(:,:)
    real(wp),             allocatable :: rest_reference(:,:)
    real(wp),             allocatable :: potential_energy(:)
    real(wp),             allocatable :: restraint_energy_0(:)
    real(wp),             allocatable :: restraint_energy_1(:)
    real(wp),             allocatable :: restraint_refcoord0(:,:)
    real(wp),             allocatable :: restraint_refcoord1(:,:)
    real(wp),             allocatable :: volume(:)
    real(wp),             allocatable :: area(:)
    real(wp),             allocatable :: after_gather (:)
    real(wp),             allocatable :: random_table (:,:)
    logical,              allocatable :: cyclic_params(:)
  end type s_remd

  ! parameters for allocatable variables
  integer, public, parameter      :: RemdReplicas    = 1
  integer, public, parameter      :: RemdUmbrellas   = 2
  integer, public, parameter      :: RemdRefatoms    = 3

  ! parameters for REMD type
  integer, public, parameter      :: RemdTemperature = 1
  integer, public, parameter      :: RemdPressure    = 2
  integer, public, parameter      :: RemdGamma       = 3
  integer, public, parameter      :: RemdRestraint   = 4

  character(*), public, parameter :: RemdTypes(4)  = (/'TEMPERATURE',&
                                                       'PRESSURE   ',&
                                                       'GAMMA      ',&
                                                       'RESTRAINT  '/)

  ! subroutines
  public :: alloc_remd
  public :: dealloc_remd
  public :: dealloc_remd_all

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    alloc_remd
  !> @brief        allocate remd information
  !! @authors      TM
  !! @param[inout] remd      : information of remd
  !! @param[in]    variable  : allocatable variable
  !! @param[in]    var_size1 : size of variables (REMD dimension)
  !! @param[in]    var_size2 : size of variables (Total number of replicas)
  !! @param[in]    var_size3 : size of variables (max number of replica)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine alloc_remd(remd, variable, var_size1, var_size2, var_size3)

    ! formal arguments
    type(s_remd),            intent(inout) :: remd
    integer,                 intent(in)    :: variable
    integer,                 intent(in)    :: var_size1
    integer,      optional,  intent(in)    :: var_size2
    integer,      optional,  intent(in)    :: var_size3

    ! local variables
    integer                  :: alloc_stat, dealloc_stat


    alloc_stat = 0
    dealloc_stat = 0

    ! allocate selected variables
    !
    select case (variable)

    case(RemdReplicas)

      if (allocated(remd%nreplicas)) then
        if (size(remd%types(:)) == var_size1) return
        deallocate(remd%types,               &
                   remd%nreplicas,           &
                   remd%cyclic_params,       &
                   remd%repid2parmsetid,     &
                   remd%repid2parmsetid_ref, &
                   remd%parmsetid2repid,     &
                   remd%parmidsets,          &
                   remd%num_criteria,        &
                   remd%num_exchanges,       &
                   remd%iparameters,         &
                   remd%dparameters,         &
                   remd%rest_function,       &
                   remd%potential_energy,    &
                   remd%restraint_energy_0,  &
                   remd%restraint_energy_1,  &
                   remd%volume,              &
                   remd%area,                &
                   remd%after_gather,        &
                   remd%random_table,        &
                   stat = dealloc_stat)
      end if

      allocate(remd%types(var_size1),                     &
               remd%nreplicas(var_size1),                 &
               remd%cyclic_params(var_size1),             &
               remd%repid2parmsetid(var_size2),           &
               remd%repid2parmsetid_ref(var_size2),       &
               remd%parmsetid2repid(var_size2),           &
               remd%parmidsets(var_size2,var_size1),      &
               remd%num_criteria(var_size2,var_size1,2),  &
               remd%num_exchanges(var_size2,var_size1,2), & 
               remd%iparameters(var_size1,var_size3),     &
               remd%dparameters(var_size1,var_size3),     &
               remd%rest_function(var_size1),             &
               remd%potential_energy(var_size2),          &
               remd%restraint_energy_0(var_size2),        &
               remd%restraint_energy_1(var_size2),        &
               remd%volume(var_size2),                    &
               remd%area(var_size2),                      &
               remd%after_gather(var_size2),              &
               remd%random_table(var_size2,var_size2),    &
               stat = alloc_stat)

      remd%types(1:var_size1)                         = 1
      remd%nreplicas(1:var_size1)                     = 0
      remd%cyclic_params(1:var_size1)                 = .false.
      remd%repid2parmsetid(1:var_size2)               = 0
      remd%repid2parmsetid_ref(1:var_size2)           = 0
      remd%parmsetid2repid(1:var_size2)               = 0
      remd%parmidsets(1:var_size2,1:var_size1)        = 0
      remd%num_criteria(1:var_size2,1:var_size1,1:2)  = 0
      remd%num_exchanges(1:var_size2,1:var_size1,1:2) = 0
      remd%iparameters(1:var_size1,1:var_size3)       = 0
      remd%dparameters(1:var_size1,1:var_size3)       = 0.0_wp
      remd%rest_function(1:var_size1)                 = 0

    case(RemdUmbrellas)

      if (allocated(remd%umbrid2numfuncs)) then
        if (size(remd%umbrid2numfuncs(:)) == var_size1) return
        deallocate(remd%umbrid2numfuncs,                  &
                   remd%umbrid2funclist,                  &
                   remd%rest_constants,                   &
                   remd%rest_reference,                   &
                   stat = dealloc_stat)
      end if

      allocate(remd%umbrid2numfuncs(var_size1),            &
               remd%umbrid2funclist(var_size1,var_size2),  &
               remd%rest_constants(var_size1,var_size3),   &
               remd%rest_reference(var_size1,var_size3),   &
               stat = alloc_stat)

      remd%umbrid2numfuncs(1:var_size1)               = 0
      remd%umbrid2funclist(1:var_size1,1:var_size2)   = 0
      remd%rest_constants(1:var_size1,1:var_size3)    = 0_wp
      remd%rest_reference(1:var_size1,1:var_size3)    = 0_wp

    case default

      call error_msg('Alloc_Replica> bad variable')

    end select


    if (alloc_stat /= 0) call error_msg_alloc
    if (dealloc_stat /= 0) call error_msg_dealloc

    return

  end subroutine alloc_remd

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    dealloc_remd
  !> @brief        deallocate remd information
  !! @authors      TM
  !! @param[inout] remd     : remd information
  !! @param[in]    variable : allocatable variable
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine dealloc_remd(remd, variable)

    ! formal arguments
    type(s_remd),            intent(inout) :: remd
    integer,                 intent(in)    :: variable

    ! local variables
    integer                  :: dealloc_stat


    dealloc_stat = 0

    ! deallocate selected variables
    !
    select case (variable)

    case (RemdReplicas)

      if (allocated(remd%types)) then
        deallocate(remd%types,               &
                   remd%nreplicas,           &
                   remd%cyclic_params,       &
                   remd%repid2parmsetid,     &
                   remd%repid2parmsetid_ref, &
                   remd%parmsetid2repid,     &
                   remd%parmidsets,          &
                   remd%num_criteria,        &
                   remd%num_exchanges,       &
                   remd%iparameters,         &
                   remd%dparameters,         &
                   remd%rest_function,       &
                   remd%potential_energy,    &
                   remd%restraint_energy_0,  &
                   remd%restraint_energy_1,  &
                   remd%volume,              &
                   remd%area,                &
                   remd%after_gather,        &
                   remd%random_table,        &
                   stat = dealloc_stat)
      end if

    case (RemdUmbrellas)

      if (allocated(remd%umbrid2funclist)) then
        deallocate(remd%umbrid2funclist,     &
                   remd%umbrid2numfuncs,     &
                   remd%rest_constants,      &
                   remd%rest_reference,      &
                   stat = dealloc_stat)
      end if

    case default

      call error_msg('Dealloc_Replica> bad variable')

    end select 


    if (dealloc_stat /= 0) call error_msg_dealloc

    return

  end subroutine dealloc_remd

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    dealloc_remd_all
  !> @brief        deallocate all remd information
  !! @authors      TM
  !! @param[inout] remd : information of remd
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine dealloc_remd_all(remd)

    ! formal arguments
    type(s_remd),         intent(inout) :: remd


    call dealloc_remd(remd, RemdReplicas)
    call dealloc_remd(remd, RemdUmbrellas)

    return

  end subroutine dealloc_remd_all

end module at_remd_str_mod
